import os.path
from flask import Blueprint, current_app, request, send_file, abort
from hgsrht.hg import HgRepository, HgCommandError, HgServerError, InvalidRef
from hgsrht.hg import lookup_rev, get_repo_tree, get_log
from io import BytesIO
from hgsrht.blueprints.api import get_user, get_repo
from srht.oauth import current_token, oauth

data = Blueprint("api_data", __name__)

def _get_repo_markers(username, reponame, accessor):
    user = get_user(username)
    repo = get_repo(user, reponame)

    results_per_page = 50
    start = request.args.get("start", 0, type=int)
    end = start + results_per_page

    with HgRepository(repo.path) as hg_repo:
        markers = accessor(hg_repo.client)
        total = len(markers)

        next_start = None
        if end < total:
            next_start = end

        return {
            "next": next_start,
            "results": [
                {
                    "name": m[0].decode(),
                    "id": m[2].decode()
                } for m in markers[start:end]
            ],
            "total": total,
            "results_per_page": results_per_page
        }

@data.route("/api/repos/<reponame>/branches", defaults={"username": None})
@data.route("/api/<username>/repos/<reponame>/branches")
@oauth("data:read")
def repo_branches_GET(username, reponame):
    return _get_repo_markers(username, reponame, lambda c: c.branches())

@data.route("/api/repos/<reponame>/bookmarks", defaults={"username": None})
@data.route("/api/<username>/repos/<reponame>/bookmarks")
@oauth("data:read")
def repo_bookmarks_GET(username, reponame):
    return _get_repo_markers(username, reponame, lambda c: c.bookmarks()[0])

@data.route("/api/repos/<reponame>/tags", defaults={"username": None})
@data.route("/api/<username>/repos/<reponame>/tags")
@oauth("data:read")
def repo_tags_GET(username, reponame):
    return _get_repo_markers(username, reponame, lambda c: c.tags())

def _rev2_to_dict(rev2):
    return {
        "id": rev2.node,
        "short_id": rev2.node_short,
        "branch": rev2.branch,
        "tags": rev2.tags,
        "author": {
            "name": rev2.author_name,
            "email": rev2.author_email
        },
        "timestamp": rev2.date,
        "message": rev2.desc,
        "phase": rev2.phase,
        "obsolete": rev2.obsolete
    }

@data.route("/api/repos/<reponame>/log", defaults={"username": None, "path": ""})
@data.route("/api/repos/<reponame>/log/<path:path>", defaults={"username": None})
@data.route("/api/<username>/repos/<reponame>/log", defaults={"path": ""})
@data.route("/api/<username>/repos/<reponame>/log/<path:path>")
@oauth("data:read")
def repo_commits_GET(username, reponame, path):
    user = get_user(username)
    repo = get_repo(user, reponame)

    results_per_page = 50
    start = request.args.get("start")
    hidden = request.args.get("hidden")

    with HgRepository(repo.path) as hg_repo:
        commits, next_id = get_log(hg_repo, start, path, results_per_page)
        return {
            "next": next_id,
            "results": [_rev2_to_dict(r) for r in commits],
            "total": -1,
            "results_per_page": results_per_page
        }

@data.route("/api/repos/<reponame>/browse", defaults={"username": None, "path": ""})
@data.route("/api/repos/<reponame>/browse/<path:path>", defaults={"username": None})
@data.route("/api/<username>/repos/<reponame>/browse", defaults={"path": ""})
@data.route("/api/<username>/repos/<reponame>/browse/<path:path>")
@oauth("data:read")
def repo_browse_GET(username, reponame, path):
    user = get_user(username)
    repo = get_repo(user, reponame)
    rev = request.args.get("rev")

    with HgRepository(repo.path) as hg_repo:
        try:
            rev = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            abort(404, str(exc))

        tree = get_repo_tree(hg_repo, rev.node, path)

        data_entries = []
        for rel_path, info in tree.items():
            full_path = os.path.join(path, rel_path)
            if info[4]:
                e = {'type': 'dir',
                     'name': rel_path,
                     'path': full_path,
                     'size': 0,
                     'perm': 0o40000}
            else:
                e = {'type': 'file',
                     'name': rel_path,
                     'path': full_path,
                     'size': info[0],
                     'perm': 0o100000 + int(info[1], 8),
                     'exec': info[2],
                     'link': info[3]}
            data_entries.append(e)
        return {
            "next": None,
            "entries": data_entries,
            "results_per_page": -1
        }

@data.route("/api/repos/<reponame>/raw/<path:path>", defaults={"username": None})
@data.route("/api/<username>/repos/<reponame>/raw/<path:path>")
@oauth("data:read")
def repo_raw_GET(username, reponame, path):
    user = get_user(username)
    repo = get_repo(user, reponame)
    rev = request.args.get("rev")

    with HgRepository(repo.path) as hg_repo:
        try:
            rev = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            abort(404, str(exc))

        try:
            full_path = os.path.join(repo.path, path)
            data = hg_repo.client.cat([full_path.encode()], rev=rev.node)
        except HgCommandError as exc:
            abort(404, exc.err.decode())

        # This is kinda lame, but it's how Mercurial seems to do it, and
        # I can't find another better way to get this info.
        is_binary = data and b'\0' in data
        fname = os.path.basename(path)
        return send_file(BytesIO(data),
                as_attachment=is_binary,
                download_name=fname,
                mimetype="text/plain" if not is_binary else None)
