import re
from flask import Blueprint, current_app, Response, render_template, abort, request
from flask import redirect, url_for
from hgsrht.hg import HgRepository, HgCommandError
from hgsrht.types import User, Repository, Redirect
from hgsrht.access import check_access, UserAccess
from srht.database import db
from srht.graphql import exec_gql
from srht.oauth import loginrequired
from srht.validation import Validation

re_changeset_hash = re.compile(r"^[0-9a-fA-F]{12,40}$")

extramanage = Blueprint('extramanage', __name__)

@extramanage.route("/<owner_name>/<repo_name>/settings/features")
@loginrequired
def manage_features(owner_name, repo_name):
    owner, repo = check_access(owner_name, repo_name, UserAccess.manage)
    if isinstance(repo, Redirect):
        return redirect(url_for(".manage_features",
            owner_name=owner_name, repo_name=repo.new_repo.name))
    return render_template("features.html", owner=owner, repo=repo)

@extramanage.route("/<owner_name>/<repo_name>/settings/features", methods=['POST'])
@loginrequired
def manage_features_POST(owner_name, repo_name):
    owner, repo = check_access(owner_name, repo_name, UserAccess.manage)
    if isinstance(repo, Redirect):
        repo = repo.new_repo
    valid = Validation(request)
    is_non_publishing = bool(valid.optional("is_non_publishing", default=''))
    if not valid.ok:
        return render_template("features.html", owner=owner, repo=repo, **valid.kwargs)

    resp = exec_gql(current_app.site, """
        mutation UpdateRepository($id: Int!, $nonPublishing: Boolean!) {
            updateRepository(id: $id, input: {nonPublishing: $nonPublishing}) { id }
        }
    """, valid=valid, id=repo.id, nonPublishing=is_non_publishing)

    if not valid.ok:
        return render_template("features.html", owner=owner, repo=repo, **valid.kwargs)
    return redirect(f"/{owner_name}/{repo_name}/settings/features")

@extramanage.route("/<owner_name>/<repo_name>/settings/histedit")
@loginrequired
def manage_histedit(owner_name, repo_name):
    owner, repo = check_access(owner_name, repo_name, UserAccess.manage)
    if isinstance(repo, Redirect):
        return redirect(url_for(".manage_histedit",
            owner_name=owner_name, repo_name=repo.new_repo.name))
    return render_template("histedit.html", owner=owner, repo=repo)

@extramanage.route("/<owner_name>/<repo_name>/settings/histedit", methods=['POST'])
@loginrequired
def manage_histedit_POST(owner_name, repo_name):
    owner, repo = check_access(owner_name, repo_name, UserAccess.manage)
    if isinstance(repo, Redirect):
        # Normally we'd redirect be we don't want to fuck up some other repo
        abort(404)
    valid = Validation(request)
    rev_id = valid.require("rev")
    if not valid.ok:
        return render_template("histedit.html", owner=owner, repo=repo, **valid.kwargs)
    rev_id = rev_id.strip()
    valid.expect(re_changeset_hash.match(rev_id), "You must pass a changeset hash.",
            field="rev")
    if not valid.ok:
        return render_template("histedit.html", owner=owner, repo=repo, **valid.kwargs)
    with HgRepository(repo.path) as hg_repo:
        strip_error = None
        strip_msg = f"changeset {rev_id} was successfully stripped!"
        try:
            hg_repo.rawcommand(b"debugstrip", rev=rev_id.encode(),
                               force=True, no_backup=True)
        except HgCommandError as exc:
            print("got hg error!")
            strip_error = exc.err.decode()
            print(strip_error)
    return render_template("histedit.html", owner=owner, repo=repo,
                           strip_msg=strip_msg, strip_error=strip_error)

