"""
This blueprint is used internally by hgsrht-shell to speed up SSH access, by
taking advantage of the database connection already established by the web app.
"""
import base64
import os
from datetime import datetime
from flask import Blueprint, request
from hgsrht.submit import trigger_changegroup_build
from hgsrht.types import User, Repository, RepoVisibility, Redirect
from hgsrht.access import get_access, UserAccess
from srht.config import cfg, get_origin
from srht.crypto import verify_request_signature
from srht.database import db
from srht.flask import csrf_bypass
from srht.graphql import exec_gql
from srht.oauth import UserType
from srht.validation import Validation

internal = Blueprint("internal", __name__)

@csrf_bypass
@internal.route("/internal/check-ssh-access", methods=["POST"])
def check_ssh_access():
    verify_request_signature(request)
    valid = Validation(request)
    path = valid.require("path")
    user_id = valid.require("user_id", cls=int)
    if not valid.ok:
        return valid.response

    # Check that the user exists.
    user = User.query.filter(User.id == user_id).one()
    if not user:
        return {
            "why": "No such user"
        }, 404

    # Check that the repository exists.
    repo = Repository.query.filter(Repository.path == path).first()
    if not repo:
        # If it's been moved, return a redirect.
        redir = Redirect.query.filter(Redirect.path == path).first()
        if redir:
            origin = get_origin("hg.sr.ht", external=True)
            repo = redir.new_repo
            # TODO: orgs
            return {
                "redirect": f'{origin}/~{repo.owner.username}/{repo.name}'
            }, 302

        # Auto-create the repository if the user is the owner of the
        # future repository.
        _path, repo_name = os.path.split(path)
        owner = os.path.basename(_path)
        if "~" + user.username != owner:
            return { }, 401

        valid = Validation({ "name": repo_name })
        resp = exec_gql(current_app.site, """
            mutation CreateRepository($name: String!) {
                createRepository(name: $name, visibility: PRIVATE) {
                    id
                    name
                }
            }
        """, valid=valid, name=name,
            description=description, visibility=visibility)
        if not valid.ok:
            return valid.response

        serve_ctx = {
            "access_grant": "rw",  # write access
            **repo.to_dict()}

    else:
        # If the user can't read the repo, refuse everything.
        access = get_access(repo, user)
        if UserAccess.read not in access:
            return {
                "why": "Unauthorized to access repository"
            }, 401
        serve_ctx = {
            "path": repo.path,
            **repo.to_dict()}

        # Take a note of whether the user has read or write access.
        access_level = "ro"
        if UserAccess.write in access:
            access_level = "rw"
        serve_ctx["access"] = access_level

    return serve_ctx, 200

@csrf_bypass
@internal.route("/internal/hook-changegroup", methods=["POST"])
def hook_changegroup():
    verify_request_signature(request)
    valid = Validation(request)
    repo_id = valid.require("repo_id")
    node_last = valid.require("node_last")
    if not valid.ok:
        return valid.response

    repo = Repository.query.get(repo_id)
    if not repo:
        return {
            "why": "Unknown repository"
        }, 404

    # Last updated time.
    repo.updated = datetime.utcnow()
    db.session.commit()

    # Trigger buildsrht builds.
    try:
        res = trigger_changegroup_build(repo, node_last)
        if res is not None:
            return res.asdict(), 200

        return {}, 200
    except Exception as ex:
        return {
            "why": str(ex)
        }, 500
