import tempfile
import hashlib
import io
import json
import os
import os.path
import re
from flask import Blueprint, Response, render_template, abort, request
from flask import redirect, url_for, send_file
from hgsrht.hg import HgRepository, HgCommandError, HgServerError, InvalidRef
from hgsrht.hg import lookup_rev, lookup_default_rev, get_repo_tree, get_log
from hgsrht.types import User, Repository, Redirect
from hgsrht.urls import clone_urls
from io import BytesIO
from markupsafe import Markup
from hgsrht.access import get_repo_or_redir
from scmsrht.formatting import get_formatted_readme, get_highlighted_file
from srht.config import cfg
from srht.markdown import sanitize
from unidiff import PatchSet

repo = Blueprint('repo', __name__)

_draft_tooltip = ("This changeset is in the draft phase and can be safely "
        "rewritten.")
_obsolete_tooltip = ("This changeset has been made obsolete by some history "
        "rewriting operation. It might have been replaced by another "
        "changeset, or removed.")

def get_readme(repo, hg_repo, root_dir, changeset, link_prefix=None):
    if repo.readme is not None:
        return Markup(sanitize(repo.readme))

    def _file_finder(name):
        try:
            bin_name = os.path.join(root_dir, name).encode('utf8')
            bin_txt = hg_repo.client.cat([bin_name], rev=changeset)
        except HgCommandError:
            return None, None

        m = hashlib.sha256()
        m.update(bin_txt)
        content_hash = m.hexdigest()
        return content_hash, bin_txt

    def _content_getter(bin_txt):
        return bin_txt.decode('utf8')

    return get_formatted_readme(
        'hg.sr.ht:hg:v2', _file_finder, _content_getter, link_prefix=link_prefix)

def get_inverse_branch_heads(hg_repo):
    branches = hg_repo.client.branches()
    return dict([(b[2].decode(), b[0].decode()) for b in branches])

def get_inverse_bookmarks(hg_repo):
    bmarks, _ = hg_repo.client.bookmarks()
    inv_bmarks = {}
    for bm_name, bm_rev, bm_node in bmarks:
        n = inv_bmarks.setdefault(bm_node.decode(), [])
        n.append(bm_name.decode())
    return inv_bmarks

def get_rev_view_data(rev2, owner, repo, inv_branch_heads=None, inv_bookmarks=None,
                      tag=None):
    rev2.short_id = rev2.node[:12]
    rev2.long_id = rev2.node
    rev = tag or rev2.node
    rev2.url = commit_rev_url(owner=owner, repo=repo, rev=rev)
    rev2.browse_url = browse_rev_url(owner=owner, repo=repo, rev=rev)
    rev2.archive_url = archive_rev_url(owner=owner, repo=repo, rev=rev)
    rev2.log_url = log_rev_url(owner=owner, repo=repo, rev=rev)
    rev2.refs = []
    rev2.aspects = []
    if inv_branch_heads and rev2.short_id in inv_branch_heads:
        rev2.refs.append({
            'name': rev2.branch, 'type': 'branch',
            'url': commit_rev_url(owner, repo, rev2.branch)})
    if inv_bookmarks and rev2.short_id in inv_bookmarks:
        rev2.refs += [
            {
                'name': b, 'type': 'bookmark',
                'url': commit_rev_url(owner, repo, b)}
            for b in inv_bookmarks[rev2.short_id]]
    rev2.refs += [
        {
            'name': t, 'type': 'tag',
            'url': commit_rev_url(owner, repo, t)}
        for t in rev2.tags]
    if rev2.obsolete:
        rev2.aspects.append('asp-obsolete')
        rev2.idextras = [
            {'name': 'obsolete', 'type': 'obsolete',
             'tooltip': _obsolete_tooltip}]
    elif rev2.phase == 'draft':
        rev2.aspects.append('asp-draft')
        rev2.idextras = [
            {'name': rev2.phase, 'type': 'phase',
             'tooltip': _draft_tooltip}]
    # TODO: add bookmarks
    rev2.parents = None # TODO
    return rev2

def get_last_3_commits(hg_repo, rev):
    revrange = f"{rev}~2..{rev}".encode()
    commits = hg_repo.log2(revrange=revrange)
    commits = list(reversed(commits))
    if len(commits) > 0 and commits[-1].rev < 0:
        commits.pop()
    return commits

def browse_rev_url(owner, repo, rev=None, path=""):
    return url_for(
        "repo.browse", owner=owner.canonical_name, repo=repo.name, rev=rev,
        path=path)

def log_rev_url(owner, repo, rev=None, path=""):
    return url_for(
        "repo.log", owner=owner.canonical_name, repo=repo.name, rev=rev,
        path=path)

def commit_rev_url(owner, repo, rev):
    return url_for(
        "repo.commit", owner=owner.canonical_name, repo=repo.name, rev=rev)

def raw_rev_url(owner, repo, rev, path):
    return url_for(
        "repo.raw_file", owner=owner.canonical_name, repo=repo.name, rev=rev,
        path=path)

def patch_rev_url(owner, repo, rev):
    return url_for(
        "repo.patch", owner=owner.canonical_name, repo=repo.name, rev=rev)

def archive_rev_url(owner, repo, rev):
    return url_for(
        "repo.archive", owner=owner.canonical_name, repo=repo.name, rev=rev)

def _highlight_file(name, content_hash, data):
    return get_highlighted_file('hg.sr.ht:hg', name, content_hash, data)

@repo.route("/<owner>/<repo>")
def summary(owner, repo):
    owner, repo = get_repo_or_redir(owner, repo)
    hg_repo = HgRepository(repo.path)

    # Check that this repo is OK before going farther.
    try:
        hg_repo.open()
    except HgServerError:
        return render_template("error-repo.html",
                owner=owner, repo=repo)

    with hg_repo:
        c_urls = clone_urls(repo)

        try:
            def_rev = lookup_default_rev(hg_repo)
            if not def_rev:
                return render_template("empty-repo.html",
                        owner=owner, repo=repo, clone_urls=c_urls)
        except HgServerError:
            return render_template("error-repo.html",
                    owner=owner, repo=repo, clone_urls=c_urls)

        tags = hg_repo.client.tags()
        # order by revision number as a cheap way to order by time.
        tags = sorted(
            filter(lambda c: c[0] != b'tip', tags),
            key=lambda c: int(c[1]), reverse=True)
        latest_tag = tags[0] if len(tags) else None

        heads = []
        def_name = def_rev.display_name
        if def_rev:
            heads.append({
                'name': def_name,
                'actions': [
                    {'name': 'browse',
                     'url': browse_rev_url(owner, repo, def_name)},
                    {'name': 'log',
                     'url': log_rev_url(owner, repo, def_name)}]})
        if latest_tag:
            latest_tag_name = latest_tag[0].decode()
            heads.append({
                'name': latest_tag_name,
                'actions': [
                    {'name': 'browse',
                     'url': browse_rev_url(owner, repo, latest_tag_name)},
                    {'name': '.tar.gz',
                     'url': archive_rev_url(owner, repo, latest_tag_name)}]})

        inv_branches = get_inverse_branch_heads(hg_repo)
        inv_bmarks = get_inverse_bookmarks(hg_repo)
        commits = [get_rev_view_data(r, owner, repo,
                        inv_branch_heads=inv_branches,
                        inv_bookmarks=inv_bmarks)
                   for r in get_last_3_commits(hg_repo, def_rev.node)]
        link_prefix = url_for(
            'repo.browse',
            owner=repo.owner,
            repo=repo.name) + '/' # Trailing slash needed
        blob_prefix = url_for(
            'repo.raw_file',
            owner=repo.owner,
            repo=repo.name,
            path="") # Trailing slash is already there because path is empty
        readme = get_readme(
            repo,
            hg_repo,
            repo.path,
            def_rev.node,
            link_prefix=[link_prefix, blob_prefix])

        if len(commits) == 0:
            return render_template(
                "empty-repo.html", view="summary", scmname="hg",
                owner=owner, repo=repo, readme=readme,
                commits=commits, clone_urls=c_urls,
                refstitle="heads", refs=heads)

        return render_template(
            "summary.html", view="summary",
            owner=owner, repo=repo, readme=readme,
            commits=commits, clone_urls=c_urls,
            refstitle="heads", refs=heads)

@repo.route("/<owner>/<repo>/browse", defaults={"path": ""})
@repo.route("/<owner>/<repo>/browse/<path:path>")
def browse(owner, repo, path):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        c_urls = clone_urls(repo)

        # Figure out the hash of the revision we want.
        rev = request.args.get("rev")
        try:
            c = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            return render_template(
                "error-invalid-ref.html", view="browse",
                owner=owner, repo=repo, error=str(exc))

        rev = c.display_name
        commit = get_rev_view_data(c, owner, repo)
        display_ref_prefix = (rev != 'default')

        # Build the breadcrumbs for the requested path.
        path_parts = [{'name': repo.name,
            'url': browse_rev_url(owner, repo, rev)}]
        parent_path_url = None
        if path:
            path_elems = path.split('/')
            cumul_path_elem = ''
            for pelem in path_elems:
                if cumul_path_elem:
                    cumul_path_elem += '/'
                cumul_path_elem += pelem
                path_parts.append({
                    'name': pelem,
                    'url': browse_rev_url(owner, repo, rev, cumul_path_elem)})

            parent_path = os.path.dirname(path)
            parent_path_url = browse_rev_url(owner, repo, rev, parent_path)

        # Get the tree listing for this revision.
        repo_tree = get_repo_tree(hg_repo, commit.long_id, path)

        # If the path we had led us to a single file, let's show it.
        if len(repo_tree) == 1 and '.' in repo_tree:
            info = repo_tree['.']
            entry = {'type': 'file',
                     'name': os.path.basename(path),
                     'size': info[0],
                     'perm': 0o100000 + int(info[1], 8),
                     'exec': info[2],
                     'link': info[3]}
            if int(info[0]) < 250*1024*1024:
                full_path = os.path.join(repo.path, path)
                rawdata = hg_repo.client.cat([full_path.encode()], rev=commit.node)
                try:
                    data = rawdata.decode()
                except:
                    data = "[can't decode contents]"
            else:
                rawdata = b"[file too big]"
                data = rawdata.decode()

            m = hashlib.sha256()
            m.update(rawdata)
            entry['hash'] = m.hexdigest()

            return render_template(
                "file.html", view='browse', owner=owner, repo=repo, ref=rev,
                display_ref_prefix=display_ref_prefix,
                path=path, path_parts=path_parts,
                entry=entry, data=data,
                raw_url=raw_rev_url(owner, repo, rev, path),
                log_url=log_rev_url(owner, repo, rev, path),
                highlight_file=_highlight_file,
                commit=commit, parent_path_url=parent_path_url,
                clone_urls=c_urls)

    # Let's show the folder/file tree.
    entries = []
    for rel_path, info in sorted(repo_tree.items(), key=lambda i: i[0]):
        full_path = os.path.join(path, rel_path)
        entry_url = browse_rev_url(owner, repo, rev, full_path)
        if info[4]:
            e = {'type': 'dir',
                 'name': rel_path,
                 'perm': 0o40000,
                 'path': full_path,
                 'browse_url': entry_url}
        else:
            e = {'type': 'file',
                 'name': rel_path,
                 'path': full_path,
                 'size': info[0],
                 'perm': 0o100000 + int(info[1], 8),
                 'exec': info[2],
                 'link': info[3],
                 'browse_url': entry_url}
        entries.append(e)

    return render_template(
        'browse.html', view='browse', scmpagetitle='browse',
        owner=owner, repo=repo, ref=rev,
        ref_tree_url=browse_rev_url(owner, repo, rev),
        ref_commit_url=commit_rev_url(owner, repo, rev),
        display_ref_prefix=display_ref_prefix,
        path=path, path_parts=path_parts, entries=entries,
        commit=commit, parent_path_url=parent_path_url,
        clone_urls=c_urls)

@repo.route("/<owner>/<repo>/raw/<path:path>")
def raw_file(owner, repo, path):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        rev = request.args.get("rev")
        try:
            c = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            abort(404)
        revid = c.node

        try:
            full_path = os.path.join(repo.path, path)
            data = hg_repo.client.cat([full_path.encode()], rev=revid)
        except HgCommandError:
            abort(404)

        # This is kinda lame, but it's how Mercurial seems to do it, and
        # I can't find another better way to get this info.
        is_binary = data and b'\0' in data
        fname = os.path.basename(path)
        return send_file(BytesIO(data),
                as_attachment=is_binary,
                download_name=fname,
                mimetype="text/plain" if not is_binary else None)

@repo.route("/<owner>/<repo>/log", defaults={"path": ""})
@repo.route("/<owner>/<repo>/log/<path:path>")
def log(owner, repo, path):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        c_urls = clone_urls(repo)

        from_rev = request.args.get("from")
        if from_rev:
            rev = from_rev
        else:
            rev = request.args.get("rev")
            try:
                c = lookup_rev(hg_repo, rev, none_on_error=False)
            except InvalidRef as exc:
                return render_template(
                    "error-invalid-ref.html", view="log",
                    owner=owner, repo=repo, error=str(exc))
            rev = c.display_name

        try:
            commits_per_page = 20
            commits, next_id = get_log(hg_repo, rev, path, commits_per_page)
            has_more = (next_id is not None)
        except HgCommandError as exc:
            return render_template(
                "error-invalid-ref.html", view="log",
                owner=owner, repo=repo, error=exc.err.decode())

        inv_branches = get_inverse_branch_heads(hg_repo)
        inv_bmarks = get_inverse_bookmarks(hg_repo)
        commits = [get_rev_view_data(r, owner, repo,
                        inv_branch_heads=inv_branches,
                        inv_bookmarks=inv_bmarks)
            for r in commits]

        next_page_url = None
        if has_more:
            query_params = {'from': next_id}
            next_page_url = url_for("repo.log",
                owner=owner.canonical_name, repo=repo.name, rev=rev, path=path,
                **query_params)

        return render_template("log.html",
                view="log", scmpagetitle="log",
                owner=owner, repo=repo, rev=rev, path=path,
                commits=commits, clone_urls=c_urls,
                has_more=has_more, next_page_url=next_page_url)

def _get_patch_file_repo_path(path):
    if path == '/dev/null':
        return ''
    return path[2:]

def _get_patch_file_url(owner, repo, revid, path):
    if not path:
        return None
    return browse_rev_url(owner, repo, revid, path)

def get_diff_data(owner, repo, hg_repo, revid):
    diff_err = None
    try:
        diff = hg_repo.client.diff(change=revid, git=True, showfunction=True)
    except HgCommandError as err:
        # Sometimes Mercurial
        if not err.out:
            raise
        diff = err.out
        diff_err = err.err.decode('utf8', 'ignore')

    with io.StringIO(diff.decode('utf8', 'ignore')) as data:
        patches = PatchSet(data)

    stats = {
        'count_files': len(patches),
        # Can use `.added` and `.removed` when we upgrade to unidiff>=0.5.5
        'count_insertions': sum([f.added for f in patches]),
        'count_deletions': sum([f.removed for f in patches])
    }

    changes = []
    for patch in patches:
        # Note that Mercurial prepends the old/new file paths with `a/`
        # and `b/` respectively so we need to strip that out, when those
        # paths are not `/dev/null` (which is the case for new or deleted
        # files).
        source_path = _get_patch_file_repo_path(patch.source_file)
        target_path = _get_patch_file_repo_path(patch.target_file)
        change = {
            'old_file': {
                'path': source_path,
                'url': _get_patch_file_url(owner, repo, revid, source_path)},
            'new_file': {
                'path': target_path,
                'url': _get_patch_file_url(owner, repo, revid, target_path)},
            'count_insertions': patch.added,
            'count_deletions': patch.removed}
        if patch.is_modified_file:
            change['status_char'] = 'M'
        elif patch.is_added_file:
            change['status_char'] = 'A'
        elif patch.is_removed_file:
            change['status_char'] = 'R'

        change['hunks'] = []
        for hunk in patch:
            hdata = {
                'old_start': hunk.source_start,
                'old_lines': hunk.source_length,
                'new_start': hunk.target_start,
                'new_lines': hunk.target_length,
                # pygit2 doesn't make the funcname/section_header separately
                # available, so we have to set the whole header to match, which
                # is \n terminated
                'header': '@@ %d,%d %d,%d @@ %s\n' % (
                    hunk.source_start,
                    hunk.source_length,
                    hunk.target_start,
                    hunk.target_length,
                    hunk.section_header,
                ),
            }
            hdata['lines'] = []
            for line in hunk:
                ldata = {'content': line.value,
                         'inserted': line.is_added,
                         'deleted': line.is_removed}
                hdata['lines'].append(ldata)

            change['hunks'].append(hdata)

        changes.append(change)

    return {'stats': stats, 'changes': changes, 'error': diff_err}

@repo.route("/<owner>/<repo>/rev/<path:rev>")
def commit(owner, repo, rev):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        c_urls = clone_urls(repo)

        try:
            commit = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            return render_template(
                "error-invalid-ref.html", view="log",
                owner=owner, repo=repo, error=str(exc))

        rev = commit.display_name
        commit = get_rev_view_data(commit, owner, repo)
        key = "hg.sr.ht:hg:commit:{}:{}:{}:v1".format(
            owner.username, repo.name, commit.node)
        try:
            diff_data = get_diff_data(owner, repo, hg_repo, commit.node)
        except HgCommandError as exc:
            return render_template("error-invalid-ref.html", view="log",
                                   owner=owner, repo=repo, error=exc.err.decode())

        return render_template("commit.html",
            view="log", scmpagetitle="commit",
            owner=owner, repo=repo, commit=commit,
            patch_url=patch_rev_url(owner, repo, rev),
            browse_url=browse_rev_url(owner, repo, rev),
            clone_urls=c_urls, **diff_data)

@repo.route("/<owner>/<repo>/rev/<path:rev>.patch")
def patch(owner, repo, rev):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        try:
            commit = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            return render_template(
                "error-invalid-ref.html", view="log",
                owner=owner, repo=repo, error=str(exc))

        rev = commit.display_name
        patch_data = hg_repo.client.export(rev.encode(), git=True)
        return send_file(BytesIO(patch_data),
                as_attachment=False,
                mimetype="text/plain")

def get_branch_info(owner, repo, hg_repo, branch):
    name, rev, node = branch
    c = get_rev_view_data(lookup_rev(hg_repo, node), owner, repo)
    c.name = name.decode()
    c.is_default = (c.name == 'default')
    return c

def get_bookmark_info(owner, repo, hg_repo, bmark):
    name, rev, node = bmark
    c = get_rev_view_data(lookup_rev(hg_repo, node), owner, repo)
    c.name = name.decode()
    return c

def get_tag_info(owner, repo, hg_repo, tag):
    name, rev, node, islocal = tag
    name = name.decode()
    c = get_rev_view_data(lookup_rev(hg_repo, node), owner, repo, tag=name)
    c.name = name
    return c

def _get_paginated_refs(refs, results_per_page=10):
    page = request.args.get("page")
    total_results = len(refs)
    total_pages = total_results // results_per_page + 1
    if total_results % results_per_page == 0:
        total_pages -= 1
    if page is not None:
        try:
            page = max(0, int(page) - 1)
        except:
            page = 0
    else:
        page = 0
    offset = page * results_per_page
    refs = refs[offset:offset+results_per_page]
    return refs, page, total_pages

@repo.route("/<owner>/<repo>/tags")
def tags(owner, repo):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        c_urls = clone_urls(repo)

        tags = hg_repo.client.tags()
        tags, page, total_pages = _get_paginated_refs(tags)
        tags = [get_tag_info(owner, repo, hg_repo, t) for t in tags]

        return render_template("tags.html",
            view="tags", scmpagetitle="tags",
            owner=owner, repo=repo, clone_urls=c_urls,
            empty_message="No tags", refs=tags,
            page=(page + 1), total_pages=total_pages)

@repo.route("/<owner>/<repo>/bookmarks")
def bookmarks(owner, repo):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        c_urls = clone_urls(repo)

        bmarks, _ = hg_repo.client.bookmarks()
        bmarks, page, total_pages = _get_paginated_refs(bmarks)
        bmarks = [get_bookmark_info(owner, repo, hg_repo, bm) for bm in bmarks]

        return render_template("bookmarks.html",
            view="bookmarks", scmpagetitle="bookmarks",
            owner=owner, repo=repo, clone_urls=c_urls,
            empty_message="No bookmarks", refs=bmarks,
            page=(page + 1), total_pages=total_pages)

@repo.route("/<owner>/<repo>/branches")
def branches(owner, repo):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        c_urls = clone_urls(repo)

        branches = hg_repo.client.branches()
        branches, page, total_pages = _get_paginated_refs(branches)
        branches = [get_branch_info(owner, repo, hg_repo, b) for b in branches]

        return render_template("branches.html",
            view="branches", scmpagetitle="branches",
            owner=owner, repo=repo, clone_urls=c_urls,
            empty_message="No branches", refs=branches,
            page=(page + 1), total_pages=total_pages)

@repo.route("/<owner>/<repo>/archive/<path:rev>.tar.gz")
def archive(owner, repo, rev):
    owner, repo = get_repo_or_redir(owner, repo)
    with HgRepository(repo.path) as hg_repo:
        try:
            commit = lookup_rev(hg_repo, rev, none_on_error=False)
        except InvalidRef as exc:
            return render_template(
                "error-invalid-ref.html", view="tags",
                owner=owner, repo=repo, error=str(exc))

        rev = commit.display_name
        rev_escaped = rev.replace('/', '_')
        basename = repo.name + "-" + rev_escaped

        with tempfile.TemporaryDirectory() as tmpdir:
            path = f"{tmpdir}/{rev_escaped}.tar.gz"
            try:
                hg_repo.client.archive(path.encode(),
                        rev=rev,
                        prefix=basename,
                        type="tgz")
            except:
                raise

            f = open(path, "rb")

        return send_file(f,
                as_attachment=True,
                download_name=f"{basename}.tar.gz",
                mimetype="application/tar+gzip")
