from datetime import datetime
from hglib.client import hgclient, revision
from hglib.error import CommandError as HgCommandError
from hglib.error import ServerError as HgServerError
from hglib.util import cmdbuilder, grouper
from srht.config import cfg
import os
import os.path
import re

_tpl_changeset2 = (b'{rev}\\0{node}\\0{tags}\\0{branch}\\0{author}'
        b'\\0{sub("\\0", "", desc)}\\0{date}\\0{phase}\\0{obsolete}\\0')

custom_hgpath = cfg("hg.sr.ht", "hg-path", None)
if custom_hgpath and os.path.exists(custom_hgpath):
    import hglib
    hglib.HGPATH = custom_hgpath
    print(f"Will use custom hg path: {custom_hgpath}")

class HgRepository:
    def __init__(self, path):
        self.path = path
        self._client = hgclient(path, encoding=None, configs=None,
                                connect=False)

    @property
    def client(self):
        return self._client

    def __enter__(self):
        self._client.__enter__()
        return self

    def __exit__(self, *args):
        self._client.__exit__(*args)

    def open(self):
        self._client.open()

    def close(self):
        self._client.close()

    def rawcommand(self, command, *args, **kwargs):
        hg_args = cmdbuilder(command, *args, **kwargs)
        return self._client.rawcommand(hg_args)

    def log2(self, revrange=None, files=[], branch=None, hidden=False,
            limit=None):
        '''a version of `hgclient.log()` that also gives you the phase
        and obsolete status of each revision.
        '''
        hg_args = cmdbuilder(b'log', template=_tpl_changeset2,
                r=revrange, b=branch, l=limit, hidden=hidden, *files)
        out = self._client.rawcommand(hg_args)
        out = out.split(b'\0')[:-1]
        return self._parserevs2(out)

    @staticmethod
    def _parserevs2(splitted):
        revs = []
        for rev in grouper(9, splitted):
            revnum = int(rev[0].decode())
            node = rev[1].decode()
            tags = rev[2].decode().split()
            branch = rev[3].decode()
            author = rev[4].decode()
            desc = rev[5].decode()
            dt = hg_date_to_datetime(rev[6])
            phase = rev[7].decode()
            obsolete = rev[8].decode()
            revs.append(revision2(
                (revnum, node, tags, branch, author, desc, dt, phase,
                obsolete)))
        return revs

def hg_date_to_timestamp(hgdate):
    return float(hgdate.split(b'.', 1)[0])

def hg_date_to_datetime(hgdate):
    posixtime = hg_date_to_timestamp(hgdate)
    return datetime.fromtimestamp(posixtime)

def get_first_line(msg):
    idx = msg.find("\n")
    if idx > 0:
        return msg[:idx]
    return msg

re_author = re.compile(r'(?P<name>.*?)\s*(\<(?P<email>.+@.+)\>)')

def split_author(author):
    m = re_author.match(author)
    if m:
        return m.group('name'), m.group('email')
    return author, None

class revision2:
    def __init__(self, vals):
        """ `vals` should be a tuple of: revnum, node, tags, branch, author,
        description, date. It can optionally also have phase and obsolete
        status.
        """
        self.rev = vals[0]
        self.node = vals[1]
        self.display_name = self.node
        self.node_short = self.node[:12]
        self.tags = vals[2]
        self.branch = vals[3]
        self.author = vals[4]
        self.author_name, self.author_email = split_author(self.author)
        self.desc = vals[5]
        self.desc_first_line = get_first_line(self.desc)
        self.date = vals[6]
        self.phase = None
        self.obsolete = None
        if len(vals) > 7:
            self.phase = vals[7]
            self.obsolete = vals[8]

    @staticmethod
    def from_rev(rev):
        return revision2((
            int(rev.rev.decode()),
            rev.node.decode(),
            rev.tags.decode().split(),
            rev.branch.decode(),
            rev.author.decode(),
            rev.desc.decode(),
            rev.date))

class InvalidRef(Exception):
    pass

def lookup_rev(hg_repo, rev, none_on_error=True):
    # If no revision is given, lookup the default one.
    if rev is None:
        return lookup_default_rev(hg_repo, none_on_error)

    # Get the log entry for the revision... if we get something, it
    # exists and we can return that.
    try:
        revs = hg_repo.log2(revrange=rev)
    except HgCommandError as exc:
        if none_on_error:
            return None
        raise InvalidRef(exc.err.decode())
    except Exception as exc:
        if none_on_error:
            return None
        raise InvalidRef(str(exc))
    if revs:
        revobj = revs[0]
        # Preserve the name the user knows this revision by.
        revobj.display_name = rev
        return revobj
    return None

def lookup_default_rev(hg_repo, none_on_error=True):
    rev = lookup_rev(hg_repo, 'tip')
    if rev:
        rev.display_name = 'tip'
        return rev
    if none_on_error:
        return None
    raise InvalidRef()

def get_repo_tree(hg_repo, revid, path):
    hg_out = hg_repo.rawcommand(b'manifest2', rev=revid, include=path)

    repo_tree = {}
    for line in hg_out.splitlines():
        perm = line[0:3].decode()
        ftypes = line[4:5].decode()
        size, fpath = line[6:].decode().split(' ', 1)
        isexec = '*' in ftypes
        islnk = '@' in ftypes
        isdir = 'd' in ftypes
        repo_tree[fpath] = (size, perm, isexec, islnk, isdir)
    return repo_tree

def get_log(hg_repo, start_commit=None, path=None, commits_per_page=20):
    files = [os.path.join(hg_repo.path, path).encode('utf8')] if path else []
    limit = commits_per_page + 1
    revrange = None
    if start_commit:
        revrange = 'reverse(ancestors({}))'.format(start_commit).encode('utf8')
    # TODO: use custom template to get merge parents, bookmarks, etc.
    commits = hg_repo.log2(revrange=revrange, limit=limit, files=files,
            hidden=True)
    # Remove the null commit if it's there.
    if len(commits) > 0 and commits[-1].rev < 0:
        commits.pop()
    # Consume our extra log entry to know if there's a next pagination ID
    # to give the caller.
    next_commit_id = None
    if len(commits) > commits_per_page:
        next_commit_id = commits[-1].node
        commits.pop()
    return commits, next_commit_id

