import os
import os.path
import requests
import stat
import sys
import tempfile
from srht.config import cfg
from werkzeug.wrappers import Request, Response

mercurial_content_type = "application/mercurial"

class HgWebShim:
    """ This is a WSGI application that just runs hgweb directly if
        the request seems to come from command-line `hg`.
    """
    def __init__(self, inner):
        self._inner = inner
        self._repos_dir = cfg("hg.sr.ht", "repos")

        try:
            from mercurial.hgweb import hgweb
            self._hgweb_class = hgweb
        except ImportError:
            print("ERROR: no mercurial installed in Python3's PYTHONPATH!")
            print("ERROR: https cloning won't work!")
            self._hgweb_class = None

    def __call__(self, environ, start_response):
        request = Request(environ)
        # We want hgweb (and, underneath it, mercurial's wire protocol server)
        # to handle this request if it looks vaguely like a repo URL, and
        # has the content-type for mercurial clients.
        if request.headers.get('Accept', '').startswith(mercurial_content_type):
            if self._hgweb_class is None:
                response = Response("wire protocol not available over HTTP",
                                    content_type="application/hg-error")
                return response(environ, start_response)
            hgweb_config = self._get_hgweb_config(request)
            if hgweb_config:
                hgweb_app = self._hgweb_class(hgweb_config)
                return hgweb_app(environ, start_response)
            else:
                response = Response(status=404)
                return response(environ, start_response)
        return self._inner(environ, start_response)

    def _get_hgweb_config(self, req):
        # Check that this repository is authorized for public https
        # cloning.
        auth_endpoint = req.url_root + 'authorize'
        headers = {'X-Original-URI': req.url}
        auth_resp = requests.get(auth_endpoint, headers=headers)
        if auth_resp.status_code != 200:
            return None

        path_bits = req.path.strip('/').split('/')
        # Only do something if it's a repo URL (`/username/repo`).
        if len(path_bits) != 2:
            return None
        user_name, repo_name = path_bits
        repo_alias = f"{user_name}/{repo_name}"
        repo_dir = os.path.join(self._repos_dir, repo_alias)
        return {repo_alias.encode(): repo_dir.encode()}
