import html
import os
import os.path
import re
import yaml
from hglib.error import CommandError as HgCommandError
from hgsrht.hg import HgRepository, revision2
from hgsrht.types import RepoVisibility
from scmsrht.submit import BuildSubmitterBase
from srht.config import cfg, get_origin
from srht.database import DbSession, db
from srht.oauth import OAuthScope
from urllib.parse import urlparse

if not hasattr(db, "session"):
    db = DbSession(cfg("hg.sr.ht", "connection-string"))
    import hgsrht.types
    db.init()

builds_sr_ht = cfg("builds.sr.ht", "origin", None)
builds_client_id = cfg("builds.sr.ht", "oauth-client-id", None)
hg_sr_ht = get_origin("hg.sr.ht", external=True)

class HgBuildSubmitter(BuildSubmitterBase):
    def __init__(self, repo, hg_repo):
        super().__init__(hg_sr_ht, 'hg', repo)
        self.hg_repo = hg_repo

    def find_manifests(self, commit):
        manifests = {}
        nodeid = commit.node.encode()
        try:
            path = os.path.join(self.repo.path, ".build.yml")
            build_yml = self.hg_repo.client.cat([path.encode()], rev=nodeid)
            manifests[".build.yml"] = build_yml
        except HgCommandError:
            pass
        if not manifests:
            builds_dir_pat = '.builds/*'.encode()
            try:
                raw_paths = self.hg_repo.rawcommand(b'files', rev=nodeid,
                        include=builds_dir_pat,
                        cwd=self.repo.path)
                paths = raw_paths.decode().strip('\n').split('\n')
            except HgCommandError:
                paths = []
            for p in paths:
                _, ext = os.path.splitext(p)
                if ext == '.yml' or ext == '.yaml':
                    name = os.path.basename(p)
                    path = os.path.join(self.repo.path, p)
                    build_yml = self.hg_repo.client.cat([path.encode()], rev=nodeid)
                    manifests[name] = build_yml
        return manifests

    def get_commit_id(self, commit):
        return commit.node

    def get_commit_note(self, commit):
        return "{}\n\n[{}]({}) &mdash; [{}](mailto:{})".format(
            "<pre>" + html.escape(commit.desc_first_line) + "</pre>",
            commit.node_short,
            "{}/{}/{}/rev/{}".format(
                hg_sr_ht,
                "~" + self.repo.owner.username,
                self.repo.name,
                commit.node),
            commit.author_name,
            commit.author_email,
        )

    def get_clone_url(self):
        origin = get_origin("hg.sr.ht", external=True)
        hg_user = cfg("git.sr.ht::dispatch", "/usr/bin/hgsrht-keys", "hg:hg").split(":")[0]
        owner_name = self.repo.owner.canonical_name
        repo_name = self.repo.name
        if self.repo.visibility == RepoVisibility.PRIVATE:
            # Use SSH URL
            origin = origin.replace("https://", "").replace("http://", "")
            return f"hg+ssh://{hg_user}@{origin}/{owner_name}/{repo_name}"
        else:
            # Use http(s) URL
            return f"hg+{origin}/{owner_name}/{repo_name}"

def trigger_changegroup_build(repo, nodeid):
    if not builds_sr_ht:
        return None

    with HgRepository(repo.path) as hg_repo:
        commits = hg_repo.client.log(revrange=nodeid.encode())
        commit = revision2.from_rev(commits[0])

        s = HgBuildSubmitter(repo, hg_repo)
        return s.submit(commit)
